<?php
/**
 * Copyright (C) 2024  Jaap Jansma (jaap.jansma@civicoop.org)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

namespace Civi\Api4\Action\BigBlueButton;

use Civi\Api4\Generic\AbstractAction;
use Civi\Api4\Generic\Result;
use Civi\Api4\Utils\ReflectionUtils;
use Civi\BigBlueButton\Call;

abstract class AbstractBBBAPi extends AbstractAction {

  /**
   * Returns the API call name. E.g. create or join.
   *
   * @return string
   */
  abstract protected function getApiCallName(): string;

  /**
   * Return the HTTP Method.
   *
   * @return string
   */
  protected function getHttpMethod(): string {
    return 'POST';
  }

  /**
   * Returns whether to only do a dry run.
   * Useful to get the call url.
   *
   * @return bool
   */
  protected function getDryRun(): bool {
    return FALSE;
  }

  /**
   * @param array $values
   *
   * @return void
   */
  public function setValues(array $values) {
    $magicProperties = static::getMagicProperties();
    foreach ($magicProperties as $name => $bool) {
      if (isset($values[$name]) && (!is_string($values[$name]) || strlen($values[$name]))) {
        $v = $values[$name];
        $this->$name = $v;
      }
    }
  }

  /**
   * @param \Civi\Api4\Generic\Result $result
   *
   * @return mixed
   */
  public function _run(Result $result) {
    $connector = bigbluebutton_connector();
    $params = $this->getParams();
    $call = $connector->call($this->getHttpMethod(), $this->getApiCallName(), $params, $this->getDryRun());
    $this->convertCallToResult($result, $call);
  }

  /**
   * @param \Civi\Api4\Generic\Result $result
   * @param \Civi\BigBlueButton\Call $call
   *
   * @return void
   */
  protected function convertCallToResult(Result $result, Call $call) {
    $result[0] = $call->getResponsArray();
    $result[0]['api_call_url'] = $call->getUrl();
  }

  /**
   * Get a list of class properties for which magic methods are supported.
   *
   * @return array
   *   List of supported properties, keyed by property name.
   *   Array(string $propertyName => bool $true).
   */
  protected static function getMagicProperties(): array {
    // Thread-local cache of class metadata. Class metadata is immutable at runtime, so this is strictly write-once. It should ideally be reused across varied test-functions.
    static $caches = [];
    $CLASS = static::CLASS;
    $cache =& $caches[$CLASS];
    if ($cache === NULL) {
      $cache = [];
      foreach (static::findStandardProperties(static::CLASS) as $property) {
        /** @var \ReflectionProperty $property */
        $cache[$property->getName()] = TRUE;
      }
    }
    return $cache;
  }

  /**
   * Get a list of standard properties which can be written+read by outside callers.
   *
   * @param string $class
   */
  private static function findStandardProperties($class): iterable {
    try {
      /** @var \ReflectionClass $clazz */
      $clazz = new \ReflectionClass($class);

      yield from [];
      foreach ($clazz->getProperties(\ReflectionProperty::IS_PROTECTED | \ReflectionProperty::IS_PUBLIC) as $property) {
        if (!$property->isStatic() && $property->getName()[0] !== '_') {
          yield $property;
        }
      }
    }
    catch (\ReflectionException $e) {
      throw new \RuntimeException(sprintf("Cannot inspect class %s.", $class));
    }
  }

}
